#include <SDL2/SDL_events.h>
#include <SDL2/SDL_render.h>
#include <SDL2/SDL_scancode.h>
#include <SDL2/SDL_surface.h>
#include <SDL2/SDL_timer.h>
#include <SDL2/SDL_video.h>
#include <stdio.h>
#include <SDL2/SDL.h>
#include <complex.h>
#include<omp.h>

#include "mandelbrot.h"
#include "mouse.h"
#include "ui.h"

int 
mandelbrot(Complex c) {
    Complex z = {0.0, 0.0};
    int i;
    
    for (i = 0; i < MAX_ITERATIONS; i++) {
        double temp_real = z.real * z.real - z.imag * z.imag + c.real;
        double temp_imag = 2 * z.real * z.imag + c.imag;
        z.real = temp_real;
        z.imag = temp_imag;
        
        if (__builtin_hypot(z.real, z.imag) > 2)
            return i;
    }
    return MAX_ITERATIONS;
} 

void 
render(App *app, ViewInfo view, int start_x, int end_x, int start_y, int end_y) {
    
    SDL_SetRenderDrawColor(app->renderer, 0, 0, 50, 255);
    SDL_RenderClear(app->renderer);

    //#pragma omp parallel for schedule(guided)
    for (int x = start_x; x < end_x; x++) {
        for (int y = start_y; y < end_y; y++) {
            double real = view.x_min + (x * (view.x_max - view.x_min)) / app->win_width;
            double imag = view.y_min + (y * (view.y_max - view.y_min)) / app->win_height;
            Complex c = {real, imag};
            
            int iterations = mandelbrot(c);

            if (iterations == MAX_ITERATIONS){
                SDL_SetRenderDrawColor(app->renderer, 0, 0, 0, 255);
            } else {
                double t = (double)iterations / MAX_ITERATIONS;
                t = 0.5 + 0.5 * cos(log(t + 0.0001) * 3.0);
                
                int r = (int)(255 * t);  
                int g = (int)(255 * t * 0.4);  
                int b = (int)(255 * t * 0.2);        
                
                SDL_SetRenderDrawColor(app->renderer, r, g, b, 255);
            }
            SDL_RenderDrawPoint(app->renderer, x, y);
        }
    }
}



int
main(int argc, char *argv[])
{
    if (SDL_Init(SDL_INIT_EVERYTHING) != 0){
        printf("error initializing SDL: %s\n", SDL_GetError());
        return 1;
    }
    
    App app = {
        SDL_CreateWindow("Mandelbrot set exploer", 
            0, 0, 
            SCREEN_WIDTH,  SCREEN_HEIGHT, 
            SDL_WINDOW_SHOWN | SDL_WINDOW_RESIZABLE | SDL_WINDOW_OPENGL), 
        SDL_CreateRenderer(app.window, 
                -1, 
                SDL_RENDERER_ACCELERATED),
        SCREEN_HEIGHT,
        SCREEN_WIDTH
            };

    ViewInfo view = {
        .x_min = -2.0,
        .x_max = 1.0,
        .y_min = -1.5,
        .y_max = 1.5,
        .zoom = 1.0
    };
    
    Mouse mouse = {
        .dragging = 0,
        .start_x = 0,
        .start_y = 0,
        .original_view = view
    };

    Uint32 frame_start;

    UI ui;
    init_ui(&ui);

    int running = 1;
    while (running)
    {

        frame_start = SDL_GetTicks();
        SDL_Event event;
        while (SDL_PollEvent(&event))
        {
            if (handle_ui_event(&ui, event, &view))
                continue;

            switch (event.type)
            {
                case SDL_QUIT:
                    running = 0;
                    break;
                case SDL_WINDOWEVENT:
                    if(event.window.event == SDL_WINDOWEVENT_RESIZED){
                        SDL_GetWindowSize(app.window, &app.win_width, &app.win_height);
                    } 
                    break;
                case SDL_KEYDOWN:
                        switch (event.key.keysym.scancode)
                        {
                            case SDL_SCANCODE_Q:
                            case SDL_SCANCODE_ESCAPE:
                                running = 0;
                                break;
                            case SDL_SCANCODE_MINUS:
                                view.x_min = view.x_min * 1.1;
                                view.x_max = view.x_max * 1.1;
                                view.y_min= view.y_min * 1.1;
                                view.y_max = view.y_max * 1.1;
                                view.zoom = view.zoom * 1.1;
                                break;
                            case SDL_SCANCODE_EQUALS:
                                view.x_min = view.x_min * 0.9;
                                view.x_max = view.x_max * 0.9;
                                view.y_min= view.y_min * 0.9;
                                view.y_max = view.y_max * 0.9;
                                view.zoom = view.zoom * 0.9;
                                break;
                        }
                default:
                    handle_mouse(event, &mouse, &view);
                    break;
            }
        }
        SDL_SetRenderDrawColor(app.renderer, 0xff, 0xff, 0xff, 0xff);
        SDL_RenderClear(app.renderer);

        //drawing happens here
        render(&app, 
                view, 
                0, app.win_width, 
                0,app.win_height);

        int msec = SDL_GetTicks() - frame_start;
        float fps = 0;
        if(msec > 0)
            fps = 1000.0 / (double) msec;

        render_ui(&ui, app.renderer, view, fps);

        SDL_RenderPresent(app.renderer);

    }
        //SDL_FreeSurface(app.surfaceA);
        SDL_DestroyRenderer(app.renderer);
        SDL_DestroyWindow(app.window);
        SDL_Quit();
}
